/* -*-c-*- 
 * Copyright Timotej Ecimovic, cic@fiz.uni-lj.si, 1998.
 *
 * This file is licensed under GNU GPL.
 *
 * It features a shared Tcl object for controling
 * the sound mixer. Purpose: to produce a neat mixer.
 *
 */
#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>
#include <fcntl.h>
#include <linux/soundcard.h>
#include <tcl.h>
#include <sys/ioctl.h>

#define Getleft(x) ((x) & 0xff)
#define Getright(x) ((x) >> 8)

char *device;
int fd=-1;
int keepalive=0;

#ifdef MIX_DEBUG
int
Mixer_ReadAll (ClientData cd, Tcl_Interp *interp,
                     int argc, char **argv )
{
  int value;

  if (!keepalive) fd=open(device, O_RDWR);
  if ( fd==-1 )
    {
      Tcl_SetResult ( interp, "Error opening mixer device.", TCL_STATIC );
      return TCL_ERROR;
    }

  ioctl (fd, SOUND_MIXER_READ_VOLUME, &value );
  printf ( "Volume:  %3d, %3d\n", Getleft(value), Getright(value));
  ioctl (fd, SOUND_MIXER_READ_BASS, &value );
  printf ( "Bass:    %3d, %3d\n", Getleft(value), Getright(value));
  ioctl (fd, SOUND_MIXER_READ_TREBLE, &value );
  printf ( "Treble:  %3d, %3d\n", Getleft(value), Getright(value));

  ioctl (fd, SOUND_MIXER_READ_SYNTH, &value ); /* FM */
  printf ( "Synth:   %3d, %3d\n", Getleft(value), Getright(value));
  ioctl (fd, SOUND_MIXER_READ_PCM, &value ); /* DSP */
  printf ( "PCM:     %3d, %3d\n", Getleft(value), Getright(value));
  ioctl (fd, SOUND_MIXER_READ_SPEAKER, &value );
  printf ( "Speaker: %3d, %3d\n", Getleft(value), Getright(value));

  ioctl (fd, SOUND_MIXER_READ_LINE, &value );
  printf ( "Line:    %3d, %3d\n", Getleft(value), Getright(value));
  ioctl (fd, SOUND_MIXER_READ_MIC, &value );
  printf ( "Mic:     %3d, %3d\n", Getleft(value), Getright(value));
  ioctl (fd, SOUND_MIXER_READ_CD, &value );
  printf ( "CD:      %3d, %3d\n", Getleft(value), Getright(value));
  ioctl (fd, SOUND_MIXER_READ_RECLEV, &value );
  printf ( "Rec.lev: %3d, %3d\n", Getleft(value), Getright(value));

  ioctl (fd, SOUND_MIXER_READ_MUTE, &value );
  printf ( "Mute: %d\n", value);
  ioctl (fd, SOUND_MIXER_READ_RECSRC, &value );
  printf ( "Rec. src.: %d\n", value);

  if (!keepalive) close(fd);
  return TCL_OK;
}
#endif

int
Mixer_Functions (ClientData cd, Tcl_Interp *interp,
                     int argc, char **argv )
{
  int value;
  int left, right;
  int write_const, read_const;

  if ( argc<2 )
    {
    error:
      Tcl_SetResult ( interp, 
		      "expected subcommand: volume, treble, bass, synth, pcm, speaker, cd, line, mic, recsrc, device, mute, stereodevs", 
		      TCL_STATIC );
      return TCL_ERROR;
    }

  if ( !strcmp (argv[1], "volume" ) )
    {
      write_const=SOUND_MIXER_WRITE_VOLUME;
      read_const=SOUND_MIXER_READ_VOLUME;
    }
  else if ( !strcmp (argv[1], "treble" ) )
    {
      write_const=SOUND_MIXER_WRITE_TREBLE;
      read_const=SOUND_MIXER_READ_TREBLE;
    }
  else if ( !strcmp (argv[1], "bass" ) )
    {
      write_const=SOUND_MIXER_WRITE_BASS;
      read_const=SOUND_MIXER_READ_BASS;
    }
  else if ( !strcmp (argv[1], "synth" ) )
    {
      write_const=SOUND_MIXER_WRITE_SYNTH;
      read_const=SOUND_MIXER_READ_SYNTH;
    }
  else if ( !strcmp (argv[1], "pcm" ) )
    {
      write_const=SOUND_MIXER_WRITE_PCM;
      read_const=SOUND_MIXER_READ_PCM;
    }
  else if ( !strcmp (argv[1], "cd" ) )
    {
      write_const=SOUND_MIXER_WRITE_CD;
      read_const=SOUND_MIXER_READ_CD;
    }
  else if ( !strcmp (argv[1], "line" ) )
    {
      write_const=SOUND_MIXER_WRITE_LINE;
      read_const=SOUND_MIXER_READ_LINE;
    }
  else if ( !strcmp (argv[1], "mic" ) )
    {
      write_const=SOUND_MIXER_WRITE_MIC;
      read_const=SOUND_MIXER_READ_MIC;
    }
  else if ( !strcmp (argv[1], "speaker" ) )
    {
      write_const=SOUND_MIXER_WRITE_SPEAKER;
      read_const=SOUND_MIXER_READ_SPEAKER;
    }
  else if ( !strcmp (argv[1], "keepalive" ) )
    {
      if ( !keepalive )
	{
	  fd=open(device, O_RDWR);
	  if ( fd==-1 )
	    {
	      Tcl_SetResult ( interp, "Error opening mixer device.", 
			      TCL_STATIC );
	      return TCL_ERROR;
	    }
	  keepalive=1;
	}
      return TCL_OK;
    }
  else if ( !strcmp (argv[1], "keepdead" ) )
    {
      if (keepalive)
	{
	  close(fd);
	  keepalive=0;
	}
      return TCL_OK;
    }
  else if ( !strcmp (argv[1], "device" ) )
    {
      if (argc==3 )
	{
	  device=(char*)realloc(device,strlen(argv[2])+1);
	  strcpy ( device, argv[2] );
	}
      else
	{
	  Tcl_SetResult ( interp, device, TCL_VOLATILE );
	}
      return TCL_OK;
    } /* device */
  else if ( !strcmp (argv[1], "stereodevs" ))
    {
      if (!keepalive) fd=open(device, O_RDWR);
      if ( fd==-1 )
	{
	  Tcl_SetResult ( interp, "Error opening mixer device.", 
			  TCL_STATIC );
	  return TCL_ERROR;
	}
      ioctl (fd, SOUND_MIXER_READ_STEREODEVS, &value );
      if (!keepalive) close(fd);
      sprintf ( interp->result, "%d", value );
      return TCL_OK;
    }
  else if ( !strcmp (argv[1], "mute" ) )
    {
      if ( argc==3 )
	{ /* Set */
	  if ( Tcl_GetInt ( interp, argv[2], &value ) != TCL_OK )
	    return TCL_ERROR;
	  if (!keepalive) fd=open(device, O_RDWR);
	  if ( fd==-1 )
	    {
	      Tcl_SetResult ( interp, "Error opening mixer device.", 
			      TCL_STATIC );
	      return TCL_ERROR;
	    }
	  ioctl (fd, SOUND_MIXER_WRITE_MUTE, &value );
	  if (!keepalive) close(fd);	  
	  return TCL_OK;
	}
      else
	{ /* Return value */
	  if (!keepalive) fd=open(device, O_RDWR);
	  if ( fd==-1 )
	    {
	      Tcl_SetResult ( interp, "Error opening mixer device.", 
			      TCL_STATIC );
	      return TCL_ERROR;
	    }
	  ioctl (fd, SOUND_MIXER_READ_RECSRC, &value );
	  if (!keepalive) close(fd);
	  sprintf ( interp->result, "%d", value );
	}
      return TCL_OK;
    }
  else if ( !strcmp (argv[1], "recsrc" ) )
    {
      if ( argc==3 )
	{ /* Set */
	  int val;
	  if ( !strcmp ( argv[2], "line" ) )
	    val=64;
	  else if ( !strcmp ( argv[2], "mic" ) )
	    val=128;
	  else if ( !strcmp ( argv[2], "cd" ) )
	    val=256;
	  else
	    {
	      Tcl_SetResult ( interp, "expected: line, mic, cd", TCL_STATIC );
	      return TCL_ERROR;
	    }
	  if (!keepalive) fd=open(device, O_RDWR);
	  if ( fd==-1 )
	    {
	      Tcl_SetResult ( interp, "Error opening mixer device.", 
			      TCL_STATIC );
	      return TCL_ERROR;
	    }
	  ioctl (fd, SOUND_MIXER_READ_RECSRC, &value );
	  value=(value&(~(64|128|256)))|val;
	  ioctl (fd, SOUND_MIXER_WRITE_RECSRC, &value );
	  if (!keepalive) close(fd);	  
	  return TCL_OK;
	}
      else
	{ /* Return value */
	  char *ret;
	  if (!keepalive) fd=open(device, O_RDWR);
	  if ( fd==-1 )
	    {
	      Tcl_SetResult ( interp, "Error opening mixer device.", 
			      TCL_STATIC );
	      return TCL_ERROR;
	    }
	  ioctl (fd, SOUND_MIXER_READ_RECSRC, &value );
	  if (!keepalive) close(fd);
	  switch(value&(64|128|256))
	    {
	    case 64:
	      ret="line";
	      break;
	    case 128:
	      ret="mic";
	      break;
	    case 256:
	      ret="cd";
	      break;
	    }
	  Tcl_SetResult ( interp, ret, TCL_STATIC );
	}
      return TCL_OK;
    } /* recsrc */
  else
    goto error;

  if ( argc==4 )
    {
      if ( Tcl_GetInt ( interp, argv[2], &left ) != TCL_OK )
	return TCL_ERROR;
      if ( Tcl_GetInt ( interp, argv[3], &right ) != TCL_OK )
	return TCL_ERROR;
      if ( left<0 || left>100 || right<0 || right>100 )
	{
	  Tcl_SetResult ( interp, "Numbers between 0 and 100 expected.", 
			  TCL_STATIC );
	  return TCL_ERROR;
	}
      value=256*right+left;
      if (!keepalive) fd=open(device, O_RDWR);
      if ( fd==-1 )
	{
	  Tcl_SetResult ( interp, "Error opening mixer device.", TCL_STATIC );
	  return TCL_ERROR;
	}
      if ( ioctl ( fd, write_const, &value ) == -1 )
        {
	  Tcl_SetResult ( interp, "Error setting volume.", TCL_STATIC );
	  if (!keepalive) close(fd);
	  return TCL_ERROR;
        }
    }
  else
    {
      if (!keepalive) fd=open(device, O_RDWR);
      if ( fd==-1 )
	{
	  Tcl_SetResult ( interp, "Error opening mixer device.", TCL_STATIC );
	  return TCL_ERROR;
	}
      ioctl (fd, read_const, &value );
      sprintf(interp->result, "%d %d", 
	      Getleft(value), Getright(value));
    }
  if (!keepalive) close(fd);
  return TCL_OK;
}

int 
Mix_Init (Tcl_Interp *interp)
{
  device=(char*)malloc(13);
  strcpy ( device, "/dev/mixer" );
    
#ifdef MIX_DEBUG
  Tcl_CreateCommand ( interp, "mix_readall", Mixer_ReadAll,
                      (ClientData)NULL, (Tcl_CmdDeleteProc*)NULL );
#endif
  Tcl_CreateCommand ( interp, "mix", Mixer_Functions,
                      (ClientData)NULL, (Tcl_CmdDeleteProc*)NULL );
  return TCL_OK;
}
